// Table for compatible characters
rawset(_G, "ManiaChars", {
    ["sonic"] = true,
	["tails"] = true,
	["knuckles"] = true,
	["fang"] = true,
	["amy"] = true,
	["metalsonic"] = true
})

rawset(_G, "NoSuper", false)

// Get done with those freeslots

freeslot(
	"S_MANIA_DUST",
	"MT_MANIA_DUST",
	"SPR_DSHR",
	"S_PLAY_WIND",
	"SPR_FAST",
	"MT_INSTASHIELD",
	"S_INSTASHIELD",
	"SPR_TWSP",
	"S_PLAY_CROUCH",
	"sfx_dpds",
	"SPR2_CRCH",
	"S_PLAY_CROUCH_START",
	"S_PLAY_CROUCH_END",
	"S_PLAY_PEELOUT_WALK",
	"S_PLAY_PEELOUT_RUN",
	"SPR2_PUSH",
	"S_PLAY_PUSH"
)

// The Sound Effect info

sfxinfo[sfx_dpds].caption = "Drop Dash Charge"

// The States and Mobjs

states[S_MANIA_DUST] = {
	sprite = SPR_DSHR,
	frame = FF_PAPERSPRITE|FF_ANIMATE|A,
	var1 = 6,
	var2 = 2,
	tics = 12,
	nextstate = S_DEATHSTATE
}

mobjinfo[MT_MANIA_DUST] = {
	doomednum = -1,
	spawnstate = S_MANIA_DUST,
	spawnhealth = 9999,
	radius = 32*FU,
	flags = MF_NOGRAVITY
}

states[S_INSTASHIELD] = {
	sprite = SPR_TWSP, 
	frame = FF_FULLBRIGHT|FF_ANIMATE, 
	tics = 7, 
	var1 = 7, 
	var2 = 1, 
	nextstate = S_NULL
}

mobjinfo[MT_INSTASHIELD] = {
	doomednum = -1,
	spawnhealth = 1,
	spawnstate = S_INSTASHIELD,
	radius = 32*FRACUNIT,
	height = 32*FRACUNIT,
	flags = MF_NOGRAVITY|MF_SCENERY|MF_NOCLIPHEIGHT
}

states[S_PLAY_WIND] = {
	sprite = SPR_FAST,
	frame = A|FF_ANIMATE|FF_PAPERSPRITE,
	tics = 10,
	var1 = 4,
	var2 = 2,
	nextstate = S_NULL
}

states[S_PLAY_CROUCH] = {
	sprite = SPR_PLAY,
	frame = SPR2_LAND,
	tics = -1,
	nextstate = S_PLAY_STND
}

states[S_PLAY_CROUCH_START] = {
	sprite = SPR_PLAY,
	frame = SPR2_CRCH|FF_ANIMATE,
	tics = 4,
	var1 = 0,
	var2 = 4,
	nextstate = S_PLAY_CROUCH
}

states[S_PLAY_CROUCH_END] = {
	sprite = SPR_PLAY,
	frame = SPR2_CRCH|FF_ANIMATE,
	tics = 4,
	var1 = 0,
	var2 = 4,
	nextstate = S_PLAY_STND
}

states[S_PLAY_PEELOUT_WALK] = {
	sprite = SPR_PLAY,
	frame = SPR2_WALK,
	tics = 4,
	var1 = 0,
	var2 = 0,
	nextstate = S_PLAY_PEELOUT_WALK
}

states[S_PLAY_PEELOUT_RUN] = {
	sprite = SPR_PLAY,
	frame = SPR2_RUN_,
	tics = 2,
	var1 = 0,
	var2 = 0,
	nextstate = S_PLAY_PEELOUT_RUN
}

states[S_PLAY_PUSH] = {
	sprite = SPR_PLAY,
	frame = SPR2_PUSH,
	tics = 32,
	var1 = 8,
	var2 = 4,
	nextstate = S_PLAY_PUSH
}

// The functions

rawset(_G, "P_IsPlayerManiafied", function(player)
	if player.mo and player.mo.valid
		local skin = player.mo.skin
		if ManiaChars[skin]
		or (skin == "s3sonic"
		or skin == "s1sonic"
		or skin == "cdsonic")
		and player.allclassics
			return true
		else
			return false
		end
	end
end)

rawset(_G, "AllClassics", function(player)
	if player.allclassics
		return true
	else
		return false
	end
end)

rawset(_G, "coollookindust", function(player)
local mo = player.mo
	if not player.powers[pw_underwater]
		if P_IsObjectOnGround(mo)
		and leveltime%10 == 0
			for i = -1, 1, 2
				local ang = i * ANG10
				local factor = FixedMul(4*mo.scale, cos(ang))
				local x = P_ReturnThrustX(mo, player.drawangle + ang, 24*mo.scale)
				local y = P_ReturnThrustY(mo, player.drawangle + ang, 24*mo.scale)
				local dust = P_SpawnMobjFromMobj(mo, -x, -y, 0, MT_MANIA_DUST)
				dust.scale = 2*$/(5/2)
				dust.angle = R_PointToAngle2(dust.x, dust.y, mo.x, mo.y)
				dust.state = S_MANIA_DUST
				if not player.fakespin
					dust.momx = FixedMul(factor, cos(dust.angle - i*ANGLE_135))
					dust.momy = FixedMul(factor, sin(dust.angle - i*ANGLE_135))
				end
			end
		end
	else
		P_SpawnSkidDust(player, 20*FRACUNIT)
	end
end)

rawset(_G, "coolspeedwindlines", function(p)
	local mo = p.mo
	local mospeed = R_PointToDist2(0, 0, R_PointToDist2(0, 0, mo.momx, mo.momy), mo.momz)
	local dist = max(p.speed / (4*FRACUNIT), 1)
	
	if leveltime%5 == 0
		for i = dist, 1, -1
			local btrail = P_SpawnMobj(p.mo.x + (P_RandomRange(-36,36) * p.mo.scale), p.mo.y + (P_RandomRange(-36,36) * p.mo.scale), p.mo.z + (p.mo.height/2) + (P_RandomRange(-20,20) * p.mo.scale), MT_SMOKE)
			btrail.state = S_PLAY_WIND
			btrail.angle = p.drawangle
			btrail.momx = 3*p.mo.momx/4
			btrail.momy = 3*p.mo.momy/4
			btrail.momz = 3*p.mo.momz/4
			btrail.colorized = true
			btrail.color = p.mo.color
			btrail.fuse = 5
			btrail.frame = $|FF_FULLBRIGHT
			btrail.blendmode = AST_ADD
			btrail.target = mo
			btrail.scale = p.mo.scale
		end
	end
end)

rawset(_G, "P_IsPlayerOnGround", function(player)
	if ((player.mo.eflags & MFE_JUSTHITFLOOR) or P_IsObjectOnGround(player.mo))
		return true
	else
		return false
	end
end)

rawset(_G, "P_CheckHeight", function(touched, toucher)
	if touched.z <= toucher.z + toucher.height
	and touched.z + touched.height >= toucher.z
		return true
	else
		return false
	end
end)

rawset(_G, "M_SpringType", function(mobj)
	if mobj and mobj.valid
		local type = mobj.type
		if type == MT_YELLOWSPRING
		or type == MT_REDSPRING
		or type == MT_BLUESPRING
			return "verti"
		elseif type == MT_YELLOWDIAG
		or type == MT_REDDIAG
		or type == MT_BLUEDIAG
			return "diag"
		elseif type == MT_YELLOWHORIZ
		or type == MT_BLUEHORIZ
		or type == MT_REDHORIZ
			return "horiz"
		end
	end
end)

rawset(_G, "M_SpawnGhost", function(player)
	local ghost = P_SpawnGhostMobj(player.mo)
	return ghost
end)

rawset(_G, "GetWallAngle", function(mo, line)
	local mobjAngle = R_PointToAngle2(line.v1.x, line.v1.y , line.v2.x, line.v2.y)
	if(#line.frontsector == #mo.subsector.sector)
		mobjAngle = $ + ANGLE_90
		mo.lineside = 0
	elseif (line.backsector ~= nil) and (#line.backsector == #mo.subsector.sector)
		mobjAngle = $ - ANGLE_90
		mo.lineside = 1
	else //just in case
		mobjAngle = $ + ANGLE_90
		mo.lineside = 0
	end
	return mobjAngle
end)

rawset(_G, "GetFloorSlopeXY", function(mo)
	local slope = mo.player.prevslope
	local zangle = slope.zangle
	local xyangle = slope.xydirection
	if (slope.flags & SL_NOPHYSICS) or (zangle == 0)
		return nil
	end
	if zangle < 0 
		xyangle = $ + ANGLE_180
	end
	if mo.eflags & MFE_VERTICALFLIP
		xyangle = $ + ANGLE_180
	end
	return xyangle
end)

rawset(_G, "thicc", function(p) -- Credit goes to NeonSRB2 for the Squash and Stretch code
	if not P_IsPlayerManiafied(p) then
		return
	end
	if p.jt == nil then
		p.jt = 0
		p.jp = false
		p.tk = false
		p.tr = false
	end
	if p.jt > 0 then
		p.jt = p.jt - 1
	end
	if p.jt < 0 then
		p.jt = p.jt + 1
	end
	if p.mo.momz < 1 then
		p.jp = false
	end
	if p.mo.state != S_PLAY_CLIMB and p.mo.eflags != p.mo.eflags | MFE_GOOWATER then
		if p.mo.momz > 0 and p.jp == false and p.mo.state != S_PLAY_FLY and p.mo.state != S_PLAY_SWIM and p.mo.state != S_PLAY_FLY_TIRED and p.mo.state != S_PLAY_WALK and p.mo.state != S_PLAY_RUN and p.mo.state != S_PLAY_WALK and p.mo.state != S_PLAY_BOUNCE_LANDING then
			p.jp = true
			p.jt = 5
		end
		if p.mo.momz > 0 and p.jt < 0 and p.mo.state != S_PLAY_FLY and p.mo.state != S_PLAY_SWIM and p.mo.state != S_PLAY_FLY_TIRED and p.mo.state != S_PLAY_WALK and p.mo.state != S_PLAY_RUN and p.mo.state != S_PLAY_WALK and p.mo.state != S_PLAY_BOUNCE_LANDING then
			p.jp = true
			p.jt = 5
		end
	elseif p.mo.eflags == p.mo.eflags | MFE_GOOWATER
		p.jp = true
	end
	if p.mo.state == S_PLAY_BOUNCE_LANDING then
		p.jt = -5
	end
	/*
	if p.pflags != p.pflags | PF_SPINNING or p.pflags == p.pflags | PF_JUMPED then
		p.sp = false
	end
	if p.pflags == p.pflags | PF_SPINNING and p.pflags != p.pflags | PF_JUMPED and p.sp == false and p.jt < 1 then
		p.sp = true
		p.jt = -5
	end
	*/
	if p.pflags != p.pflags | PF_THOKKED then
		p.tk = false
	end
	if p.pflags == p.pflags | PF_THOKKED and p.tk == false then
		p.tk = true
		p.jt = 5
	end
	if p.mo.state != S_PLAY_FLY_TIRED then
		p.tr = false
	end
	if p.mo.state == S_PLAY_FLY_TIRED and p.tr == false then
		p.tr = true
		p.jt = 5
	end
	p.maths = p.jt*FRACUNIT
	p.maths = p.maths / 10
	p.mo.spriteyscale = p.maths + FRACUNIT
	
	p.maths = p.jt*p.spinheight
	if p.mo.state != S_PLAY_ROLL then
		p.maths = p.jt*p.height
	end
	p.maths = p.maths / 20
	p.mo.spriteyoffset = -1*p.maths
	
	p.maths = p.jt*FRACUNIT
	p.maths = p.maths / 10
	p.maths = p.maths*-1
	p.mo.spritexscale = p.maths + FRACUNIT
end)

// HPU rq

rawset(_G, "GetWallAngle", function(mo, line)
	local mobjAngle = R_PointToAngle2(line.v1.x, line.v1.y , line.v2.x, line.v2.y)
	if(#line.frontsector == #mo.subsector.sector)
		mobjAngle = $ + ANGLE_90
		mo.lineside = 0
	elseif (line.backsector ~= nil) and (#line.backsector == #mo.subsector.sector)
		mobjAngle = $ - ANGLE_90
		mo.lineside = 1
	else //just in case
		mobjAngle = $ + ANGLE_90
		mo.lineside = 0
	end
	return mobjAngle
end)

rawset(_G, "GetFloorSlopeXY", function(mo)
	local slope = mo.player.prevslope
	local zangle = slope.zangle
	local xyangle = slope.xydirection
	if (slope.flags & SL_NOPHYSICS) or (zangle == 0)
		return nil
	end
	if zangle < 0 
		xyangle = $ + ANGLE_180
	end
	if mo.eflags & MFE_VERTICALFLIP
		xyangle = $ + ANGLE_180
	end
	return xyangle
end)

// Commands

COM_AddCommand("dropdash", function(player) //Can dropdash
	if player.candropdash //If on
		player.candropdash = false //Make it off
		CONS_Printf(player, "No dropdash for you") //CONS_Printf text
	else //If off
		player.candropdash = true //Make it on
		CONS_Printf(player, "Mania time") //CONS_Printf Text
	end
end) //End command

COM_AddCommand("dropdashonspin", function(player) //Dropdash on spin?
	if player.dpspin //If on
		player.dpspin = false //Make it off
		CONS_Printf(player, "Dropdash is now on Jump") //CONS_Printf text
	else //If off
		player.dpspin = true //Make it on
		CONS_Printf(player, "Dropdash is now on Spin") //CONS_Printf Text
	end
end) //End command

COM_AddCommand("dropdashanim", function(player) //Dropdash Animation
	if player.dodpanim //If on
		player.dodpanim = false //Make it off
		CONS_Printf(player, "No more dropdash animation") //CONS_Printf text
	else //If off
		player.dodpanim = true //Make it on
		CONS_Printf(player, "Bad Dropdash animation for you") //CONS_Printf Text
	end
end) //End command

COM_AddCommand("dpsquake", function(player) //Earthquake after dropdash
	if player.dpsquake //If on
		player.dpsquake = false //Make it off
		CONS_Printf(player, "No earthquake after a dropdash while super") //CONS_Printf text
	else //If off
		player.dpsquake = true //Make it on
		CONS_Printf(player, "There will now be an earthquake after a dropdash while super") //CONS_Printf text
		CONS_Printf(player, "Warning you, if you have epilepsy, I recommend you turn this off") //CONS_Printf warning
	end
end) //End command

COM_AddCommand("speedcap", function(player) // Should running be capped?
	if player.speedcap //If on
		player.speedcap = false //Make it off
		CONS_Printf(player, "Run to your heart's content!") //CONS_Printf text
	else //If off
		player.speedcap = true //Make it on
		CONS_Printf(player, "There's now a speedcap on your movement!") //CONS_Printf text
	end
end) //End command

/*COM_AddCommand("IsManiaChar", function(player)
	if ManiaChars[player.mo.skin]
		print "Ye"
	else
		print "nop"
	end
end)

COM_AddCommand("AddManiaChar", function(player, arg)
	if arg
		if arg == "self"
			table.insert(ManiaChars, skins[player.mo.skin].name)
		else
			local skin = tostring(arg)
			if skins[skin].valid
				table.insert(ManiaChars, skins[skin].name)
			else
				CONS_Printf(player, "Skin not valid!")
			end
		end
	else
		CONS_Printf(player, "No arg inputted!")
	end
end)

COM_AddCommand("ThrustPlayer", function()
	for player in players.iterate()
		P_InstaThrust(player.mo, player.mo.angle, 300*FRACUNIT)
	end
end)*/

COM_AddCommand("momentum", function(player)
	if player.momentum
		player.momentum = false
		CONS_Printf(player, "Momentum is now off")
	else
		player.momentum = true
		CONS_Printf(player, "Momentum is now on")
	end
end)

COM_AddCommand("anglestand", function(player)
	if player.anglestand
		player.anglestand = false
		CONS_Printf(player, "Anglestand is now off")
	else
		player.anglestand = true
		CONS_Printf(player, "Anglestand is now on")
	end
end)

COM_AddCommand("airwalk", function(player)
	if player.airwalk
		player.airwalk = false
		CONS_Printf(player, "Airwalk is now off")
	else
		player.airwalk = true
		CONS_Printf(player, "Airwalk is now on")
	end
end)

COM_AddCommand("fallbackcam", function(player)
	if player.fallback
		player.fallback = false
		CONS_Printf(player, "Fall back is now off")
	else
		player.fallback = true
		CONS_Printf(player, "Fall back is now on")
	end
end)

COM_AddCommand("classicphys", function(player)
	if player.classic
		player.classic = false
		CONS_Printf(player, "Classic is now off")
	else
		player.classic = true
		CONS_Printf(player, "Classic is now on")
	end
end)

COM_AddCommand("maniashoes", function(player)
	if player.shoes
		player.shoes = false
		CONS_Printf(player, "AI Shoes are now off")
	else
		player.shoes = true
		CONS_Printf(player, "AI Shoes are now on")
	end
end)

COM_AddCommand("CDSprings", function(player)
	if player.cd
		player.cd = false
		CONS_Printf(player, "CD is now off")
	else
		player.cd = true
		CONS_Printf(player, "CD is now on")
	end
end)

COM_AddCommand("SolidSprings", function(player)
	if player.sp
		player.sp = false
		CONS_Printf(player, "Solid Springs are now off")
	else
		player.sp = true
		CONS_Printf(player, "Solid Springs are now on")
	end
end)

COM_AddCommand("Reroll", function(player)
	if player.reroll
		player.reroll = false
		CONS_Printf(player, "Reroll is now off")
	else
		player.reroll = true
		CONS_Printf(player, "Reroll is now on")
	end
end)

COM_AddCommand("SRB2Spindash", function(player)
	if player.crouching
		player.crouching = false
		CONS_Printf(player, "SRB2 Spindash is now on")
	else
		player.crouching = true
		CONS_Printf(player, "SRB2 Spindash is now off")
	end
end)

COM_AddCommand("Peelout", function(player)
	if player.peelout
		player.peelout = false
		CONS_Printf(player, "Peelout is now off")
	else
		player.peelout = true
		CONS_Printf(player, "Peelout is now on")
	end
end)

COM_AddCommand("Instashield", function(player)
	if player.insta
		player.insta = false
		CONS_Printf(player, "Instashield is now off")
	else
		player.insta = true
		CONS_Printf(player, "Instashield is now on")
	end
end)

COM_AddCommand("Instashieldonspin", function(player)
	if player.instas
		player.instas = false
		CONS_Printf(player, "Instashield is now on Jump")
	else
		player.instas = true
		CONS_Printf(player, "Instashield is now on Spin")
	end
end)

COM_AddCommand("rollafteruse", function(player)
	if player.rollafteruse
		player.rollafteruse = false
		CONS_Printf(player, "Roll after use is now off")
	else
		player.rollafteruse = true
		CONS_Printf(player, "Roll after use is now on")
	end
end)

COM_AddCommand("ISRebound", function(player)
	if player.instar
		player.instar = false
		CONS_Printf(player, "Instashield Rebound is now off")
	else
		player.instar = true
		CONS_Printf(player, "Instashield Rebound is now on")
	end
end)

COM_AddCommand("timeout", function(player)
	if player.timeout
		player.timeout = false
		CONS_Printf(player, "Timeout is now off")
	else
		player.timeout = true
		CONS_Printf(player, "Timeout is now on")
	end
end)

COM_AddCommand("cdtimeout", function(player)
	if player.cdtimeout
		player.cdtimeout = false
		CONS_Printf(player, "CD Timeout is now off")
	else
		player.cdtimeout = true
		CONS_Printf(player, "CD Timeout is now on")
	end
end)

COM_AddCommand("rollangle", function(player)
	if player.rollangle
		player.rollangle = false
		CONS_Printf(player, "Roll angle is now off")
	else
		player.rollangle = true
		CONS_Printf(player, "Roll angle is now on")
	end
end)

COM_AddCommand("rolloutspindash", function(player)
	if player.rollout
		player.rollout = false
		CONS_Printf(player, "Rollout spindash is now off")
	else
		player.rollout = true
		CONS_Printf(player, "Rollout spindash is now on")
	end
end)

COM_AddCommand("doloops", function(player)
	if player.loops
		player.loops = false
		CONS_Printf(player, "You can now do loops!")
	else
		player.loops = true
		CONS_Printf(player, "Loops have been disabled!")
	end
end)

COM_AddCommand("speedeffects", function(player)
	if player.speedeffect
		player.speedeffect = false
		CONS_Printf(player, "Speed effects are off!")
	else
		player.speedeffect = true
		CONS_Printf(player, "Speed effects are now on!")
	end
end)

COM_AddCommand("waterrun", function(player)
	if player.waterrun
		player.waterrun = false
		CONS_Printf(player, "The player can't run on water anymore!")
	else
		player.waterrun = true
		CONS_Printf(player, "The player can now run on water! (Only at high speeds)")
	end
end)

COM_AddCommand("playerstretch", function(player)
	if player.stretch
		player.stretch = false
		CONS_Printf(player, "Stretch is now off!")
	else
		player.stretch = true
		CONS_Printf(player, "Stretch is now on!")
	end
end)

COM_AddCommand("runonwatereffect", function(player)
	if player.waterruneffect
		player.waterruneffect = false
		CONS_Printf(player, "Running on water is now boring!")
	else
		player.waterruneffect = true
		CONS_Printf(player, "Cool effects now happen when you run on water!")
	end
end)

COM_AddCommand("AllClassics", function(player)
	if player.allclassics
		player.allclassics = false
		CONS_Printf(player, "Other Classics (such as S3 Sonic) now are restored to their own physics.")
	else
		player.allclassics = true
		CONS_Printf(player, "Other Classics (such as S3 Sonic) now get Mania physics.")
	end
end)

COM_AddCommand("movestack", function(player)
	if player.movestack == false
		player.movestack = true
		CONS_Printf(player, "If your dropdash is on spin, you can thok with jump!")
	else
		player.movestack = false
		CONS_Printf(player, "You can't have the dropdash and thok at the same time.")
	end
end)

// Setup all the options

addHook("PreThinkFrame", function()
	for player in players.iterate()
		if player.momentum == nil
			player.momentum = true
		end
		if player.anglestand == nil
			player.anglestand = true
		end
		if player.airwalk == nil
			player.airwalk = true
		end
		if player.fallback == nil
			player.fallback = true
		end
		if player.shoes == nil
			player.shoes = true
		end
		if player.classic == nil
			player.classic = true
		end
		if player.cd == nil
			player.cd = true
		end
		if player.sp == nil
			player.sp = false
		end
		if player.reroll == nil
			player.reroll = false
		end
		if player.crouching == nil
			player.crouching = true
		end
		if player.peelout == nil
			player.peelout = false
		end
		if player.insta == nil
			player.insta = false
		end
		if player.instas == nil
			player.instas = true
		end
		if player.instar == nil
			player.instar = false
		end
		if player.rollafteruse == nil
			player.rollafteruse = false
		end
		if player.timeout == nil
			player.timeout = false
		end
		if player.rollangle == nil
			player.rollangle = true
		end
		if player.rollout == nil
			player.rollout = false
		end
		if player.loops == nil
			player.loops = false
		end
		if player.speedeffect == nil
			player.speedeffect = false
		end
		if player.peeloutanim == nil
			player.peeloutanim = true
		end
		if player.waterrun == nil
			player.waterrun = true
		end
		if player.stretch == nil
			player.stretch = false
		end
		if player.waterruneffect == nil
			player.waterruneffect = true
		end
		if player.push == nil
			player.push = true
		end
		if player.allclassics == nil
			player.allclassics = false
		end
		if player.smusic == nil
			player.smusic = "mania"
		end
		if player.speedchange
			player.speedchange = true
		end
		if player.movestack == nil
			player.movestack = true
		end
		if player.speedcap == nil
			player.speedcap = true
		end
		if player.jumptics == nil
			player.jumptics = 0
		end
		if (player.cmd.buttons & BT_JUMP)
			player.jumptics = $ + 1
		else
			player.jumptics = 0
		end
	end
end)

// The Tables

rawset(_G, "MSStartup", function(player)
	if player.menu == nil
	or not player.menu
		player.menu = {
			selected = "continue",
			changing = false
		}
		
		CONS_Printf(player, "\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\nThe best and only SRB2 Engine, The " + "\x87" + "Classic" + "\x84" + " Engine!")
		CONS_Printf(player, "Classic Engine made by \x8Ekirb\x80")
	end
end)

addHook("PlayerThink", MSStartup)